@@ -34,6 +34,10 @@
 
 #include "exelf.h"
 
+#define EXIT_KERNEL_ERROR  -1
+#define EXIT_LOGFILE_ERROR -2
+#define EXIT_SYSDATA_ERROR -3
+
 #define EXMFILE "/proc/exmap"
 #define BUF_SIZE 256
 
@@ -125,13 +129,22 @@ typedef struct
   GHashTable * files;
   GHashTable * map;
   gchar      * cmd;
+  gboolean     load_symbols;
 } memory_t;
 
 typedef struct
 {
   memory_t      m;
   const gchar * name;
-  
+  esym_t      * esym;
+}symbol_t;
+
+typedef struct
+{
+  memory_t      m;
+  const gchar * name;
+
+  GList * syms;
 }section_t;
 
 typedef struct
@@ -148,6 +161,19 @@ page_start (address_t addr)
   return addr & ~(getpagesize()-1);
 }
 
+typedef struct
+{
+  pid_t       * pids_arg;
+  int           pids_count;
+  char       ** pids_names;
+  guint         detail;
+  gboolean      all_pids;
+  guint         sort;
+  guint         threshold;
+  format_t      format;
+  const char  * log;
+} args_t;
+
 static void
 vma_free (gpointer p)
 {
@@ -260,7 +286,7 @@ process_vma (gpointer key, gpointer value, gpointer data)
 
           /* store the file name in the cmd field, it comes handy later */
           fm->m.cmd = g_strdup (v->file);
-          fm->elf = elf_load (v->file);
+          fm->elf = elf_load (v->file, m->load_symbols);
           fm->m.pg_count = m->pg_count;
           
           g_hash_table_insert (m->files, g_strdup (v->file), fm);
@@ -391,39 +417,84 @@ process_files (gpointer key, gpointer value, gpointer data)
   while (l)
     {
       esec_t * s = (esec_t*) l->data;
-      section_t * sec = g_malloc0(sizeof(section_t));
+      section_t * sec = g_malloc0 (sizeof (section_t));
 
       sec->name = elf_section_name (s);
       sec->m.pg_count = f->m.pg_count;
 
+      /* create any symbol data structures */
+      GList * k = elf_section_symbols (s);
+      while (k)
+        {
+          esym_t    * esbl = k->data;
+          symbol_t  * sbl = g_malloc0 (sizeof (symbol_t));
+          
+          sbl->esym = esbl;
+          sbl->name = elf_symbol_name (esbl);
+          sbl->m.pg_count = f->m.pg_count;
+          
+          sec->syms = g_list_prepend (sec->syms, sbl);
+          k = k->next;
+        }
+
+      if (sec->syms)
+        sec->syms = g_list_reverse (sec->syms);
+      
       if (!sec->m.pg_count)
         g_warning ("Missing pg_count hash");
       
       f->sects = g_list_prepend (f->sects, sec);
 
-      GList * k = f->vma;
+      k = f->vma;
       while (k)
         {
           vma_t * v = (vma_t*)k->data;
           address_t start = v->start;
           address_t end   = v->end;
 
-          if ((start != end) &&
-              elf_section_intersect (s, &start, &end, v->offset))
+          if (start != end)
             {
-              process_pages_in_range (&(sec->m), v->pages, start, end);
+              /* NB: the order here matters, as elf_section_interset()
+               * modifies the start and end values
+               */
+              if (sec->syms)
+                {
+                  /* This is a symtable; process the symbols
+                   *
+                   * TODO -- is there a more efficient way of doing this ?
+                   */
+                  GList * syms = sec->syms;
+                  while (syms)
+                    {
+                      symbol_t  * sbl = syms->data;
+                      address_t   start2 = start;
+                      address_t   end2 = end;
+                      
+                      if (elf_symbol_intersect (sbl->esym,
+                                                &start2, &end2, v->offset))
+                        {
+                          process_pages_in_range (&(sbl->m),
+                                                  v->pages, start2, end2);
+                        }
+                      
+                      syms = syms->next;
+                    }
+                }
+              else if (elf_section_intersect (s, &start, &end, v->offset))
+                {
+                  process_pages_in_range (&(sec->m), v->pages, start, end);
+                }
             }
           
           k = k->next;
         }
-      
+
       l = l->next;
     }
 }
 
-/* TODO refactor these print fuctions to share code */
 static guint
-print_section_info (section_t * s, format_t format)
+print_symbol_info (FILE * of, symbol_t * s, format_t format)
 {
   if (!s || !s->m.vm)
     return 0;
@@ -432,7 +503,7 @@ print_section_info (section_t * s, format_t format)
   
   if (format == FORMAT_HUMAN)
     {
-      g_print ("   %-29s: VM%8u %s, M%8u %s, S%8u %s\n",
+      fprintf (of,"      %-26s: VM%8u %s, M%8u %s, S%8u %s\n",
                s->name,
                SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
                SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
@@ -442,7 +513,7 @@ print_section_info (section_t * s, format_t format)
     }
   else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
     {
-      g_print (",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
+      fprintf (of,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
                s->name, 
                s->m.vm,
                s->m.mapped,
@@ -453,9 +524,66 @@ print_section_info (section_t * s, format_t format)
 
   return lines;
 }
-     
+
+/* TODO refactor these print fuctions to share code */
 static guint
-print_file_info (file_t * f, format_t format, guint detail)
+print_section_info (FILE * of, section_t * s, format_t format)
+{
+  if (!s)
+    return 0;
+
+  guint lines = 0;
+
+  if (s->m.vm)
+    {
+      if (format == FORMAT_HUMAN)
+        {
+          fprintf (of,"   %-29s: VM%8u %s, M%8u %s, S%8u %s\n",
+                   s->name,
+                   SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
+                   SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
+                   SCALE_V(s->m.sole),   SCALE_U(s->m.sole));
+      
+          lines++;
+        }
+      else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
+        {
+          fprintf (of,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
+                   s->name, 
+                   s->m.vm,
+                   s->m.mapped,
+                   s->m.sole);
+
+          lines++;
+        }
+    }
+  else if (s->syms)
+    {
+      if (format == FORMAT_HUMAN)
+        {
+          fprintf (of,"   %-29s:\n", s->name);
+          lines++;
+        }
+      else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
+        {
+          fprintf (of,",,,,,,,,,,,,\"%s\",0,0,0\n", s->name);
+          lines++;
+        }
+    }
+  
+  GList * l = s->syms;
+  while (l)
+    {
+      symbol_t * sbl = l->data;
+      lines += print_symbol_info (of, sbl, format);
+      l = l->next;
+    }
+  
+  return lines;
+}
+
+static guint
+print_file_info (FILE * of, file_t * f, format_t format, guint detail)
 {
   if (!f)
     return 0;
@@ -467,34 +595,34 @@ print_file_info (file_t * f, format_t format, guint detail)
     {
       if (detail & DETAIL_ELF_INFO)
         {
-          g_print ("  ----------------------------------"
+          fprintf (of,"  ----------------------------------"
                    "---------------------------------------\n");
           lines++;
         }
 
-      g_print ("  %-30s: VM%8u %s, M%8u %s, S%8u %s ",
+      fprintf (of,"  %-30s: VM%8u %s, M%8u %s, S%8u %s ",
                cmd,
                SCALE_V(f->m.vm),     SCALE_U(f->m.vm),
                SCALE_V(f->m.mapped), SCALE_U(f->m.mapped),
                SCALE_V(f->m.sole),   SCALE_U(f->m.sole));
-#ifdef DEBUG
-      elf_print_info (f->elf);
+#if 0
+      elf_print_dbg_info (f->elf);
 #endif
       if (detail & DETAIL_ELF_INFO)
         {
           lines += 2;
-          g_print ("\n  ----------------------------------"
+          fprintf (of,"\n  ----------------------------------"
                    "---------------------------------------\n");
         }
       else
         {
           lines++;
-          g_print ("\n");
+          fprintf (of,"\n");
         }
     }
   else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
     {
-      g_print (",,,,,,,,\"%s\",%u,%u,%u\n",
+      fprintf (of,",,,,,,,,\"%s\",%u,%u,%u\n",
                f->m.cmd, 
                f->m.vm,
                f->m.mapped,
@@ -511,7 +639,7 @@ print_file_info (file_t * f, format_t format, guint detail)
       while (l)
         {
           section_t * s= (section_t*)l->data;
-          lines += print_section_info (s, format);
+          lines += print_section_info (of, s, format);
           l = l->next;
         }
     }
@@ -538,7 +666,7 @@ load_pids ()
     }
 
   g_dir_close (dir);
-  
+
   return pids;
 }
 
@@ -716,7 +844,8 @@ load_map (pid_t pid)
 static memory_t *
 load_pid_data (pid_t pid,
                GHashTable * map_hash,
-               GHashTable * pg_count)
+               GHashTable * pg_count,
+               args_t * args)
 {
   FILE     * fexmp = NULL;
   FILE     * fcmd = NULL;
@@ -747,6 +876,53 @@ load_pid_data (pid_t pid,
             *n = 0;
 
           mem_data->cmd = g_strdup (buf);
+
+          /* see if this process is among named processes given to us on the
+           * cmd line
+           */
+          if (args->pids_names && args->pids_names[0])
+            {
+              /* Unfortunately, g_path_basename() does not work if the
+               * cmd line has any parameters, so we extract the process name
+               * ourselves.
+               * 
+               * -- this assumes no spaces in file names
+               */
+              gchar * cmd2 = g_strdup (mem_data->cmd);
+              gchar * c;
+              gchar * name;
+
+              if ((c = strrchr (cmd2, ' ')))
+                *c = 0;
+
+              if ((c = strrchr (cmd2, '/')))
+                name = c+1;
+              else
+                name = cmd2;
+
+              char ** p = args->pids_names;
+
+              while (*p)
+                {
+                  if (!strcmp (name, *p))
+                    {
+                      /* add space for the new pid
+                       * 2 == 1 for new pid 1 for NULL terminator
+                       */
+                      args->pids_arg = g_realloc (args->pids_arg,
+                                                  args->pids_count + 2);
+                      
+                      args->pids_arg[args->pids_count] = pid;
+                      args->pids_count++;
+                      args->pids_arg[args->pids_count] = 0;
+                    }
+                  
+                  p++;
+                }
+
+              g_free (cmd2);
+            }
+          
         }
       fclose (fcmd);
     }
@@ -758,7 +934,7 @@ load_pid_data (pid_t pid,
   if (!fexmp)
     {
       g_warning ("Could not open [%s]. Is exmap.ko loaded ?\n", EXMFILE);
-      exit (1);
+      exit (EXIT_KERNEL_ERROR);
     }
 
   if (EOF == fprintf (fexmp, "%d", pid))
@@ -930,7 +1106,7 @@ sort_pid_list (gconstpointer a, gconstpointer b, gpointer data)
  * file_info indicates if per-file information is to be printed as well
  */
 static guint
-print_results (memory_t * mem_data, guint detail, guint threshold,
+print_results (FILE * of, memory_t * mem_data, guint detail, guint threshold,
                format_t format, int sort)
 {
   guint lines = 0;
@@ -943,28 +1119,28 @@ print_results (memory_t * mem_data, guint detail, guint threshold,
 
   if (format == FORMAT_HUMAN)
     {
-      g_print ("Process %d", mem_data->pid);
+      fprintf (of,"Process %d", mem_data->pid);
 
       if (mem_data->cmd)
-        g_print (" [%s]\n", mem_data->cmd);
+        fprintf (of," [%s]\n", mem_data->cmd);
       else
-        g_print ("\n");
+        fprintf (of,"\n");
   
-      g_print (" Virtual memory  : %9u %s\n",
+      fprintf (of," Virtual memory  : %9u %s\n",
                SCALE_V(mem_data->vm), SCALE_U(mem_data->vm));
-      g_print (" Effective VM    : %9u %s\n",
+      fprintf (of," Effective VM    : %9u %s\n",
                SCALE_V(mem_data->evm), SCALE_U(mem_data->evm));
-      g_print (" Heap            : %9u %s\n",
+      fprintf (of," Heap            : %9u %s\n",
                SCALE_V(mem_data->heap), SCALE_U(mem_data->heap));
-      g_print (" Stack           : %9u %s\n",
+      fprintf (of," Stack           : %9u %s\n",
                SCALE_V(mem_data->stack), SCALE_U(mem_data->stack));
-      g_print (" vdso            : %9u %s\n",
+      fprintf (of," vdso            : %9u %s\n",
                SCALE_V(mem_data->vdso), SCALE_U(mem_data->vdso));
-      g_print (" Mapped          : %9u %s\n",
+      fprintf (of," Mapped          : %9u %s\n",
                SCALE_V(mem_data->mapped), SCALE_U(mem_data->mapped));
-      g_print (" Effective mapped: %9u %s\n",
+      fprintf (of," Effective mapped: %9u %s\n",
                SCALE_V(mem_data->emapped), SCALE_U(mem_data->emapped));
-      g_print (" Sole use        : %9u %s\n",
+      fprintf (of," Sole use        : %9u %s\n",
                SCALE_V(mem_data->sole), SCALE_U(mem_data->sole));
 
       lines += 9;
@@ -972,14 +1148,14 @@ print_results (memory_t * mem_data, guint detail, guint threshold,
       if (detail & DETAIL_FILE_INFO)
         {
           lines += 2;
-          g_print ("\n Per file memory use\n");
+          fprintf (of,"\n Per file memory use\n");
         }
     }
   else if (format == FORMAT_CSV || format == FORMAT_CSV_SUMS)
     {
       lines++;
       
-      g_print ("%u,\"%s\",%u,%u,%u,%u,%u,%u,%u,%u\n",
+      fprintf (of,"%u,\"%s\",%u,%u,%u,%u,%u,%u,%u,%u\n",
                mem_data->pid,
                mem_data->cmd ? mem_data->cmd : "???",
                mem_data->vm,
@@ -1002,7 +1178,7 @@ print_results (memory_t * mem_data, guint detail, guint threshold,
           l = g_list_sort_with_data (l, sort_pid_list, GINT_TO_POINTER (sort));
           while (l)
             {
-              lines += print_file_info ((file_t*)l->data, format, detail);
+              lines += print_file_info (of, (file_t*)l->data, format, detail);
               l = l->next;
             }
 
@@ -1012,7 +1188,7 @@ print_results (memory_t * mem_data, guint detail, guint threshold,
 
   if (format == FORMAT_HUMAN)
     {
-      g_print ("\n========================================"
+      fprintf (of,"\n========================================"
                "=======================================\n\n");
     }
 
@@ -1042,30 +1218,30 @@ usage (const char * option)
       fprintf (f, "\nUnknown option %s.\n", option);
     }
   else
-    fprintf (f, "Exmap is a simple console front end to the exmap.ko module\n");
+    fprintf (f, "\nThis is the command-line front-end to Exmap.\n");
   
-  fprintf (f, "\nUsage: exmap [options] pid [pid ...]\n\n");
-  fprintf (f, "  --no-file-info        : Only summary information for each process\n");
-  fprintf (f, "  --no-elf-info         : No breakdown per elf sections in each file\n");
-  fprintf (f, "  --all-pids            : Information for all accessible pids\n");
+  fprintf (f, "\nUsage: exmap [options] [pid|name ...]\n\n");
+  fprintf (f, "Options:\n");
+  fprintf (f, "  --no-file-info        : Only summary for each process (implies --no-elf-info).\n");
+  fprintf (f, "  --no-elf-info         : No breakdown per elf sections in each file.\n");
+  fprintf (f, "  --all-pids            : Information for all accessible pids; if not used,\n"
+              "                          at least one process must be specified.\n");
   fprintf (f, "  --vm-threshold=bytes  : Information only for processes above VM threshold\n");
   fprintf (f, "  --sort-by=what        : vm|sole|heap|stack|mapped|none (default: none)\n");
   fprintf (f, "  --sort-type=which     : desc|asc (default: desc)\n");
   fprintf (f, "  --format=which        : human|csv|csv+sums (default: human)\n");
+  fprintf (f, "  --log=file            : print results to file, not stdout\n");
   fprintf (f, "  --version             : Program version\n");
   fprintf (f, "  --module-version      : Version of required kernel module\n");
   fprintf (f, "  --help\n\n");
-}
+  fprintf (f, "If --all-pids is not given, at least one process has to be specified,\n");
+  fprintf (f, "either by a pid or command name. Command names must not include path. \n");
+  fprintf (f, "Specifying a process by name is equivalent to giving pids for all running \n");
+  fprintf (f, "processes of that name.\n\n");
 
-typedef struct
-{
-  pid_t      * pids_arg;
-  guint        detail;
-  gboolean     all_pids;
-  guint        sort;
-  guint        threshold;
-  format_t     format;
-} args_t;
+  fprintf (f, "If neither --no-file-info nor --no-elf-info is given then for any \n");
+  fprintf (f, "processes specified explicitely symbol information is printed where available.\n\n");
+}
 
 static gboolean
 parse_args (int argc, char ** argv, args_t * args)
@@ -1147,6 +1323,13 @@ parse_args (int argc, char ** argv, args_t * args)
           continue;
         }
 
+      if (!strncmp (argv[i], "--log",5))
+        {
+          args->log = argv[i] + 6;
+
+          continue;
+        }
+      
       if (!strcmp (argv[i], "--help"))
         {
           usage (NULL);
@@ -1171,20 +1354,42 @@ parse_args (int argc, char ** argv, args_t * args)
       return FALSE;
     }
   
-  args->pids_arg = g_malloc0((argc - i + 1) * sizeof(pid_t));
+  args->pids_arg   = g_malloc0((argc - i + 1) * sizeof(pid_t));
+  args->pids_names = g_malloc0((argc - i + 1) * sizeof(const char*));
+  args->pids_count = 0;
   j = 0;
   
   for (; !args->all_pids && i < argc; ++i)
     {
-      args->pids_arg[j] = atoi (argv[i]);
+      args->pids_arg[args->pids_count] = atoi (argv[i]);
 
-      if(args->pids_arg[j])
-        j++;
+      if(args->pids_arg[args->pids_count])
+        args->pids_count++;
+      else
+        {
+          /* assume this is a process name */
+          args->pids_names[j] = argv[i];
+          j++;
+        }
     }
 
   return TRUE;
 }
 
+static gboolean
+is_pid_present (pid_t * pids, pid_t pid)
+{
+  while (pids && *pids)
+    {
+      if (*pids == pid)
+        return TRUE;
+
+      pids++;
+    }
+
+  return FALSE;
+}
+
 int
 main (int argc, char ** argv)
 {
@@ -1195,24 +1400,38 @@ main (int argc, char ** argv)
   GList      * l;
   gint         retval = 0;
   guint        lines = 0;
-  
+  FILE *       of = stdout;
   args_t       args;
-  args.pids_arg = NULL;
-  args.detail = (DETAIL_FILE_INFO | DETAIL_ELF_INFO);
-  args.all_pids  = FALSE;
-  args.sort = 0;
-  args.threshold = 0;
-  args.format = FORMAT_HUMAN;
+  time_t       now;
+  struct tm    t;
+  char         date_stamp[30];
+  
+  args.pids_arg   = NULL;
+  args.pids_names = NULL;
+  args.detail     = (DETAIL_FILE_INFO | DETAIL_ELF_INFO);
+  args.all_pids   = FALSE;
+  args.sort       = 0;
+  args.threshold  = 0;
+  args.format     = FORMAT_HUMAN;
+  args.log        = NULL;
 
   if (! parse_args (argc, argv, &args))
     {
       retval = -1;
       goto finish;
     }
+
+  if (args.log && (of = fopen (args.log, "w")) == NULL)
+    {
+      g_warning ("Could not open putput file [%s]", args.log);
+      retval = EXIT_LOGFILE_ERROR;
+      goto finish;
+    }
   
   if (! (pids_sys = load_pids ()))
     {
-      retval = -1;
+      g_warning ("There was a problem loading system data");
+      retval = EXIT_SYSDATA_ERROR;
       goto finish;
     }
 
@@ -1230,19 +1449,24 @@ main (int argc, char ** argv)
       if (!h)
         continue;
       
-      if ((m = load_pid_data (p, h, pg_count)))
-        mem_list = g_list_prepend (mem_list, m);
-
+      if ((m = load_pid_data (p, h, pg_count, &args)))
+        {
+          m->load_symbols = is_pid_present (args.pids_arg, p);
+          mem_list = g_list_prepend (mem_list, m);
+        }
+      
       l = l->next;
     }
 
   /* work out how much memory is shared between processes
    * and how much between files.
    */
+  
   l = mem_list;
   while (l)
     {
       memory_t * m = (memory_t *) l->data;
+
       g_hash_table_foreach (m->map, process_vma, m);
       g_hash_table_foreach (m->files, process_files, m);
 
@@ -1256,22 +1480,30 @@ main (int argc, char ** argv)
                                         GINT_TO_POINTER (args.sort));
     }
 
+  /* Print a time stamp into the log */
+  now = time (NULL);
+  localtime_r (&now, &t);
+  strftime (date_stamp, sizeof (date_stamp),
+            "%a %Y %m %d %H:%M:%S", &t);
+                      
+  fprintf (of,"--- Exmap log from %s ---\n\n", date_stamp);
+  
   /* output header */
   if (args.format == FORMAT_CSV || args.format == FORMAT_CSV_SUMS)
     {
-      g_print ("pid,command,VM,effective VM,heap,stack,vdso,mapped,sole");
+      fprintf (of,"pid,command,VM,effective VM,heap,stack,vdso,mapped,sole");
 
       if (args.detail & DETAIL_FILE_INFO)
         {
-          g_print (",file,file-VM,file-mapped,file-solefile");
+          fprintf (of,",file,file-VM,file-mapped,file-solefile");
 
           if (args.detail & DETAIL_ELF_INFO)
-            g_print (",section,section-VM,section-mapped,section-sole\n");
+            fprintf (of,",section,section-VM,section-mapped,section-sole\n");
           else
-            g_print ("\n");
+            fprintf (of,"\n");
         }
       else
-        g_print ("\n");
+        fprintf (of,"\n");
     }
 
   /* output info for individula pids */
@@ -1280,7 +1512,7 @@ main (int argc, char ** argv)
       l = mem_list;
       while (l)
         {
-          lines += print_results ((memory_t*)l->data,
+          lines += print_results (of, (memory_t*)l->data,
                                   args.detail, args.threshold, args.format, args.sort);
           l = l->next;
         }
@@ -1291,8 +1523,11 @@ main (int argc, char ** argv)
       while (*pid)
         {
           l = g_list_find_custom (mem_list, GINT_TO_POINTER (*pid), find_pid);
-          lines += print_results ((memory_t*)l->data,
-                                  args.detail, args.threshold, args.format, args.sort);
+
+          if (l)
+            lines += print_results (of, (memory_t*)l->data,
+                                    args.detail, args.threshold,
+                                    args.format, args.sort);
           pid++;
         }
     }
@@ -1303,12 +1538,12 @@ main (int argc, char ** argv)
       int i;
       char c = 'A' + 2;
 
-      g_print ("Totals,\"(F9 to recalculate)\""); /* in place of pid, cmd */
+      fprintf (of,"Totals,\"(F9 to recalculate)\""); /* in place of pid, cmd */
       
       for (i = 0; i < CSV_COLUMNS - 2; ++i, ++c)
-        g_print (",\"=sum(%c2:%c%d)\"", c, c, lines+1);
+        fprintf (of,",\"=sum(%c2:%c%d)\"", c, c, lines+1);
 
-      g_print ("\n");
+      fprintf (of,"\n");
 
       lines++;
 
@@ -1321,6 +1556,9 @@ main (int argc, char ** argv)
                  lines);
     }
 
+  /* put a mark in, se we know that the process finished normally */
+  fprintf (of,"--- Exmap log ends here ---\n");
+  
  finish:
   /* clean up */
   if (pg_count)
@@ -1344,6 +1582,8 @@ main (int argc, char ** argv)
 
   if (args.pids_arg)
     g_free (args.pids_arg);
+
+  fclose (of);
   
   return retval;
 }
